@@ -20,16 +20,12 @@ where
     Graph::EdgeData: Clone,
 {
     let mut result = Vec::new();
-    // Using a vector might be faster here. We anyways only pop from one side.
-    let mut nodes: LinkedList<_> = graph.node_indices().collect();
-    // TODO this is not optimal. The bfs recreates a vector of all nodes all the time.
-    // Instead of doing that, the bfs could reuse the order vector.
-    // Then, an offset would need to be used for the subgraph indices.
+    let mut nodes: Vec<_> = graph.node_indices().collect();
     let mut visited = vec![false; graph.node_count()];
     let mut bfs = PreOrderUndirectedBfs::new_without_start(graph);
 
     while !nodes.is_empty() {
-        let start = nodes.pop_front().unwrap();
+        let start = nodes.pop().unwrap();
         if visited[start.as_usize()] {
             continue;
         }
@@ -233,24 +229,32 @@ where
     Graph::NodeData: Clone,
     Graph::EdgeData: Clone,
 {
-    let mut result = Vec::new();
-    let mut extracted_nodes = vec![Graph::OptionalNodeIndex::new_none(); graph.node_count()];
+    // invert node mapping in linear time
+    let mut root_node_map = vec![usize::MAX; graph.node_count()];
+    let mut subgraph_node_indices = Vec::new();
+    for (node_index, root_node) in node_mapping.iter().enumerate() {
+        let node_index = Graph::NodeIndex::from(node_index);
+        let subgraph_index = root_node_map[root_node.as_usize()];
+        if subgraph_index == usize::MAX {
+            root_node_map[root_node.as_usize()] = subgraph_node_indices.len();
+            subgraph_node_indices.push(vec![node_index]);
+        } else {
+            subgraph_node_indices[subgraph_index].push(node_index);
+        }
+    }
+
+    // extract subgraphs
     let mut id_map = Vec::new();
+    let mut extracted_nodes = vec![Graph::OptionalNodeIndex::new_none(); graph.node_count()];
 
-    for node in graph.node_indices() {
-        if !extracted_nodes[node.as_usize()].is_valid() {
-            //println!("Processing {:?}", node);
-
-            let root_node = node_mapping[node.as_usize()];
+    subgraph_node_indices
+        .iter()
+        .map(|subgraph_node_indices| {
             let mut subgraph = Graph::default();
             id_map.clear();
+            let root_node = node_mapping[subgraph_node_indices.first().unwrap().as_usize()];
 
-            for node in graph
-                .node_indices()
-                .skip(node.as_usize())
-                .filter(|n| node_mapping[n.as_usize()] == root_node)
-            {
-                //println!("Adding node {:?}", node);
+            for &node in subgraph_node_indices {
                 let subgraph_node = subgraph.add_node(graph.node_data(node).clone());
                 extracted_nodes[node.as_usize()] = subgraph_node.into();
                 id_map.push(node);
@@ -271,11 +275,9 @@ where
                 }
             }
 
-            result.push(subgraph);
-        }
-    }
-
-    result
+            subgraph
+        })
+        .collect()
 }
 
 /// Returns true if the given graph is a cycle.
@@ -300,44 +302,43 @@ mod tests {
         let n2 = graph.add_node(2);
         let n3 = graph.add_node(3);
         let n4 = graph.add_node(4);
-        let e0 = graph.add_edge(n0, n1, 10);
-        let e1 = graph.add_edge(n1, n2, 11);
-        let e15 = graph.add_edge(n1, n2, 115);
-        let e2 = graph.add_edge(n2, n3, 12);
-        let e3 = graph.add_edge(n3, n4, 13);
-        let e4 = graph.add_edge(n4, n0, 14);
-        let e45 = graph.add_edge(n4, n0, 145);
-        let e5 = graph.add_edge(n1, n0, 15);
-        let e6 = graph.add_edge(n2, n1, 16);
-        let e7 = graph.add_edge(n3, n2, 17);
-        let e8 = graph.add_edge(n4, n3, 18);
-        let e9 = graph.add_edge(n0, n4, 19);
-        let e10 = graph.add_edge(n2, n2, 20);
+        let _e0 = graph.add_edge(n0, n1, 10);
+        let _e1 = graph.add_edge(n1, n2, 11);
+        let _e15 = graph.add_edge(n1, n2, 115);
+        let _e2 = graph.add_edge(n2, n3, 12);
+        let _e3 = graph.add_edge(n3, n4, 13);
+        let _e4 = graph.add_edge(n4, n0, 14);
+        let _e45 = graph.add_edge(n4, n0, 145);
+        let _e5 = graph.add_edge(n1, n0, 15);
+        let _e6 = graph.add_edge(n2, n1, 16);
+        let _e7 = graph.add_edge(n3, n2, 17);
+        let _e8 = graph.add_edge(n4, n3, 18);
+        let _e9 = graph.add_edge(n0, n4, 19);
+        let _e10 = graph.add_edge(n2, n2, 20);
         let result = decompose_weakly_connected_components(&graph);
         debug_assert_eq!(result.len(), 1);
         let result = result.first().unwrap();
         debug_assert_eq!(result.node_count(), graph.node_count());
         debug_assert_eq!(result.edge_count(), graph.edge_count());
 
-        debug_assert_eq!(result.node_data(n0), &0);
-        debug_assert_eq!(result.node_data(n1), &4);
-        debug_assert_eq!(result.node_data(n2), &1);
-        debug_assert_eq!(result.node_data(n3), &3);
-        debug_assert_eq!(result.node_data(n4), &2);
+        let mut node_data: Vec<_> = graph
+            .node_indices()
+            .map(|i| result.node_data(i))
+            .copied()
+            .collect();
+        node_data.sort_unstable();
+        debug_assert_eq!(node_data, vec![0, 1, 2, 3, 4]);
 
-        debug_assert_eq!(result.edge_data(e0), &145);
-        debug_assert_eq!(result.edge_data(e1), &14);
-        debug_assert_eq!(result.edge_data(e15), &19);
-        debug_assert_eq!(result.edge_data(e2), &15);
-        debug_assert_eq!(result.edge_data(e3), &10);
-        debug_assert_eq!(result.edge_data(e4), &13);
-        debug_assert_eq!(result.edge_data(e45), &18);
-        debug_assert_eq!(result.edge_data(e5), &20);
-        debug_assert_eq!(result.edge_data(e6), &16);
-        debug_assert_eq!(result.edge_data(e7), &12);
-        debug_assert_eq!(result.edge_data(e8), &17);
-        debug_assert_eq!(result.edge_data(e9), &115);
-        debug_assert_eq!(result.edge_data(e10), &11);
+        let mut edge_data: Vec<_> = graph
+            .edge_indices()
+            .map(|i| result.edge_data(i))
+            .copied()
+            .collect();
+        edge_data.sort_unstable();
+        debug_assert_eq!(
+            edge_data,
+            vec![10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 115, 145]
+        );
     }
 
     #[test]
@@ -348,36 +349,36 @@ mod tests {
         let n2 = graph.add_node(2);
         let n3 = graph.add_node(3);
         let n4 = graph.add_node(4);
-        let e0 = graph.add_edge(n0, n1, 10);
-        let e1 = graph.add_edge(n1, n2, 11);
-        let e2 = graph.add_edge(n2, n3, 12);
-        let e3 = graph.add_edge(n3, n4, 13);
-        let e4 = graph.add_edge(n1, n0, 15);
-        let e5 = graph.add_edge(n3, n2, 17);
-        let e6 = graph.add_edge(n4, n3, 18);
-        let e7 = graph.add_edge(n2, n2, 20);
-        let e8 = graph.add_edge(n2, n2, 21);
+        let _e0 = graph.add_edge(n0, n1, 10);
+        let _e1 = graph.add_edge(n1, n2, 11);
+        let _e2 = graph.add_edge(n2, n3, 12);
+        let _e3 = graph.add_edge(n3, n4, 13);
+        let _e4 = graph.add_edge(n1, n0, 15);
+        let _e5 = graph.add_edge(n3, n2, 17);
+        let _e6 = graph.add_edge(n4, n3, 18);
+        let _e7 = graph.add_edge(n2, n2, 20);
+        let _e8 = graph.add_edge(n2, n2, 21);
         let result = decompose_weakly_connected_components(&graph);
         debug_assert_eq!(result.len(), 1);
         let result = result.first().unwrap();
         debug_assert_eq!(result.node_count(), graph.node_count());
         debug_assert_eq!(result.edge_count(), graph.edge_count());
 
-        debug_assert_eq!(result.node_data(n0), &0);
-        debug_assert_eq!(result.node_data(n1), &1);
-        debug_assert_eq!(result.node_data(n2), &2);
-        debug_assert_eq!(result.node_data(n3), &3);
-        debug_assert_eq!(result.node_data(n4), &4);
+        let mut node_data: Vec<_> = graph
+            .node_indices()
+            .map(|i| result.node_data(i))
+            .copied()
+            .collect();
+        node_data.sort_unstable();
+        debug_assert_eq!(node_data, vec![0, 1, 2, 3, 4]);
 
-        debug_assert_eq!(result.edge_data(e0), &15);
-        debug_assert_eq!(result.edge_data(e1), &10);
-        debug_assert_eq!(result.edge_data(e2), &21);
-        debug_assert_eq!(result.edge_data(e3), &20);
-        debug_assert_eq!(result.edge_data(e4), &11);
-        debug_assert_eq!(result.edge_data(e5), &17);
-        debug_assert_eq!(result.edge_data(e6), &12);
-        debug_assert_eq!(result.edge_data(e7), &18);
-        debug_assert_eq!(result.edge_data(e8), &13);
+        let mut edge_data: Vec<_> = graph
+            .edge_indices()
+            .map(|i| result.edge_data(i))
+            .copied()
+            .collect();
+        edge_data.sort_unstable();
+        debug_assert_eq!(edge_data, vec![10, 11, 12, 13, 15, 17, 18, 20, 21]);
     }
 
     #[test]
@@ -396,9 +397,9 @@ mod tests {
         graph.add_edge(n4, n5, 13);
         let result = decompose_weakly_connected_components(&graph);
         debug_assert_eq!(result.len(), 3);
-        let first = result[0].clone();
+        let first = result[2].clone();
         let second = result[1].clone();
-        let third = result[2].clone();
+        let third = result[0].clone();
         debug_assert_eq!(first.node_count(), 1);
         debug_assert_eq!(first.edge_count(), 1);
         debug_assert_eq!(second.node_count(), 2);
@@ -407,17 +408,17 @@ mod tests {
         debug_assert_eq!(third.edge_count(), 2);
 
         debug_assert_eq!(first.node_data(0.into()), &0);
-        debug_assert_eq!(second.node_data(0.into()), &1);
-        debug_assert_eq!(second.node_data(1.into()), &2);
-        debug_assert_eq!(third.node_data(0.into()), &3);
+        debug_assert_eq!(second.node_data(1.into()), &1);
+        debug_assert_eq!(second.node_data(0.into()), &2);
+        debug_assert_eq!(third.node_data(2.into()), &3);
         debug_assert_eq!(third.node_data(1.into()), &4);
-        debug_assert_eq!(third.node_data(2.into()), &5);
+        debug_assert_eq!(third.node_data(0.into()), &5);
 
         debug_assert_eq!(first.edge_data(0.into()), &10);
         debug_assert_eq!(second.edge_data(0.into()), &115);
         debug_assert_eq!(second.edge_data(1.into()), &11);
-        debug_assert_eq!(third.edge_data(0.into()), &12);
-        debug_assert_eq!(third.edge_data(1.into()), &13);
+        debug_assert_eq!(third.edge_data(1.into()), &12);
+        debug_assert_eq!(third.edge_data(0.into()), &13);
     }
 
     #[test]
@@ -435,42 +436,42 @@ mod tests {
         let n2 = graph.add_node(2);
         let n3 = graph.add_node(3);
         let n4 = graph.add_node(4);
-        let e0 = graph.add_edge(n0, n1, 10);
-        let e05 = graph.add_edge(n0, n1, 105);
-        let e1 = graph.add_edge(n1, n2, 11);
-        let e2 = graph.add_edge(n2, n3, 12);
-        let e3 = graph.add_edge(n3, n4, 13);
-        let e4 = graph.add_edge(n4, n1, 14);
-        let e5 = graph.add_edge(n1, n4, 15);
-        let e6 = graph.add_edge(n2, n1, 16);
-        let e7 = graph.add_edge(n3, n2, 17);
-        let e8 = graph.add_edge(n4, n3, 18);
-        let e9 = graph.add_edge(n0, n4, 19);
-        let e10 = graph.add_edge(n2, n2, 20);
+        let _e0 = graph.add_edge(n0, n1, 10);
+        let _e05 = graph.add_edge(n0, n1, 105);
+        let _e1 = graph.add_edge(n1, n2, 11);
+        let _e2 = graph.add_edge(n2, n3, 12);
+        let _e3 = graph.add_edge(n3, n4, 13);
+        let _e4 = graph.add_edge(n4, n1, 14);
+        let _e5 = graph.add_edge(n1, n4, 15);
+        let _e6 = graph.add_edge(n2, n1, 16);
+        let _e7 = graph.add_edge(n3, n2, 17);
+        let _e8 = graph.add_edge(n4, n3, 18);
+        let _e9 = graph.add_edge(n0, n4, 19);
+        let _e10 = graph.add_edge(n2, n2, 20);
         let result = decompose_weakly_connected_components(&graph);
         debug_assert_eq!(result.len(), 1);
         let result = result.first().unwrap();
         debug_assert_eq!(result.node_count(), graph.node_count());
         debug_assert_eq!(result.edge_count(), graph.edge_count());
 
-        debug_assert_eq!(result.node_data(n0), &0);
-        debug_assert_eq!(result.node_data(n1), &4);
-        debug_assert_eq!(result.node_data(n2), &1);
-        debug_assert_eq!(result.node_data(n3), &3);
-        debug_assert_eq!(result.node_data(n4), &2);
+        let mut node_data: Vec<_> = graph
+            .node_indices()
+            .map(|i| result.node_data(i))
+            .copied()
+            .collect();
+        node_data.sort_unstable();
+        debug_assert_eq!(node_data, vec![0, 1, 2, 3, 4]);
 
-        debug_assert_eq!(result.edge_data(e0), &19);
-        debug_assert_eq!(result.edge_data(e05), &15);
-        debug_assert_eq!(result.edge_data(e1), &14);
-        debug_assert_eq!(result.edge_data(e2), &105);
-        debug_assert_eq!(result.edge_data(e3), &10);
-        debug_assert_eq!(result.edge_data(e4), &13);
-        debug_assert_eq!(result.edge_data(e5), &18);
-        debug_assert_eq!(result.edge_data(e6), &20);
-        debug_assert_eq!(result.edge_data(e7), &16);
-        debug_assert_eq!(result.edge_data(e8), &12);
-        debug_assert_eq!(result.edge_data(e9), &17);
-        debug_assert_eq!(result.edge_data(e10), &11);
+        let mut edge_data: Vec<_> = graph
+            .edge_indices()
+            .map(|i| result.edge_data(i))
+            .copied()
+            .collect();
+        edge_data.sort_unstable();
+        debug_assert_eq!(
+            edge_data,
+            vec![10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 105]
+        );
     }
 
     #[test]
@@ -481,42 +482,42 @@ mod tests {
         let n2 = graph.add_node(2);
         let n3 = graph.add_node(3);
         let n4 = graph.add_node(4);
-        let e0 = graph.add_edge(n4, n1, 10);
-        let e1 = graph.add_edge(n1, n2, 11);
-        let e2 = graph.add_edge(n2, n3, 12);
-        let e3 = graph.add_edge(n3, n4, 13);
-        let e4 = graph.add_edge(n4, n0, 14);
-        let e5 = graph.add_edge(n1, n0, 15);
-        let e55 = graph.add_edge(n1, n0, 155);
-        let e6 = graph.add_edge(n2, n1, 16);
-        let e7 = graph.add_edge(n3, n2, 17);
-        let e8 = graph.add_edge(n4, n3, 18);
-        let e9 = graph.add_edge(n1, n4, 19);
-        let e10 = graph.add_edge(n2, n2, 20);
+        let _e0 = graph.add_edge(n4, n1, 10);
+        let _e1 = graph.add_edge(n1, n2, 11);
+        let _e2 = graph.add_edge(n2, n3, 12);
+        let _e3 = graph.add_edge(n3, n4, 13);
+        let _e4 = graph.add_edge(n4, n0, 14);
+        let _e5 = graph.add_edge(n1, n0, 15);
+        let _e55 = graph.add_edge(n1, n0, 155);
+        let _e6 = graph.add_edge(n2, n1, 16);
+        let _e7 = graph.add_edge(n3, n2, 17);
+        let _e8 = graph.add_edge(n4, n3, 18);
+        let _e9 = graph.add_edge(n1, n4, 19);
+        let _e10 = graph.add_edge(n2, n2, 20);
         let result = decompose_weakly_connected_components(&graph);
         debug_assert_eq!(result.len(), 1);
         let result = result.first().unwrap();
         debug_assert_eq!(result.node_count(), graph.node_count());
         debug_assert_eq!(result.edge_count(), graph.edge_count());
 
-        debug_assert_eq!(result.node_data(n0), &0);
-        debug_assert_eq!(result.node_data(n1), &1);
-        debug_assert_eq!(result.node_data(n2), &4);
-        debug_assert_eq!(result.node_data(n3), &2);
-        debug_assert_eq!(result.node_data(n4), &3);
+        let mut node_data: Vec<_> = graph
+            .node_indices()
+            .map(|i| result.node_data(i))
+            .copied()
+            .collect();
+        node_data.sort_unstable();
+        debug_assert_eq!(node_data, vec![0, 1, 2, 3, 4]);
 
-        debug_assert_eq!(result.edge_data(e0), &155);
-        debug_assert_eq!(result.edge_data(e1), &15);
-        debug_assert_eq!(result.edge_data(e2), &14);
-        debug_assert_eq!(result.edge_data(e3), &10);
-        debug_assert_eq!(result.edge_data(e4), &19);
-        debug_assert_eq!(result.edge_data(e5), &20);
-        debug_assert_eq!(result.edge_data(e55), &16);
-        debug_assert_eq!(result.edge_data(e6), &11);
-        debug_assert_eq!(result.edge_data(e7), &17);
-        debug_assert_eq!(result.edge_data(e8), &13);
-        debug_assert_eq!(result.edge_data(e9), &18);
-        debug_assert_eq!(result.edge_data(e10), &12);
+        let mut edge_data: Vec<_> = graph
+            .edge_indices()
+            .map(|i| result.edge_data(i))
+            .copied()
+            .collect();
+        edge_data.sort_unstable();
+        debug_assert_eq!(
+            edge_data,
+            vec![10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 155]
+        );
     }
 
     #[test]
